/* pl.rmalinowski.gwtreflect.client.reflect.Method.java 
 Copyright (C) 1998, 1999, 2001, 2005  Free Software Foundation, Inc.

 This file is part of GNU Classpath.

 GNU Classpath is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 GNU Classpath is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with GNU Classpath; see the file COPYING.  If not, write to the
 Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 02110-1301 USA.

 Linking this library statically or dynamically with other modules is
 making a combined work based on this library.  Thus, the terms and
 conditions of the GNU General Public License cover the whole
 combination.

 As a special exception, the copyright holders of this library give you
 permission to link this library with independent modules to produce an
 executable, regardless of the license terms of these independent
 modules, and to copy and distribute the resulting executable under
 terms of your choice, provided that you also meet, for each linked
 independent module, the terms and conditions of the license of that
 module.  An independent module is a module which is not derived from
 or based on this library.  If you modify this library, you may extend
 this exception to your version of the library, but you are not
 obligated to do so.  If you do not wish to do so, delete this
 exception statement from your version. */

package pl.rmalinowski.gwtreflect.client.reflect;

import pl.rmalinowski.gwtreflect.client.reflect.types.Void;
import pl.rmalinowski.gwtreflect.client.utils.JavaToJSArrayUtil;

import com.google.gwt.core.client.JavaScriptObject;

public abstract class Method 
    extends AccessibleObject 
    implements Member {

    private final String name;

    private final Class declaringClass;

    private final int modifiers;

    private final String[] parameterSignatures;

    private final String[] exceptionSignatures;

    private final String returnSignature;

    private Class[] parameterTypes;

    private Class[] exceptionTypes;

    private Class returnType;

    protected abstract Object nativeInvoke(Object obj, JavaScriptObject argsArray);

    // private final boolean isSignatured;

    public Method(Class declaringClass, String name, int modifiers) {
        this(declaringClass, name, modifiers, Void.SIGNATURE, new String[] {}, new String[] {});
    }

    public Method(Class declaringClass, String name, int modifiers, String returnSignature) {
        this(declaringClass, name, modifiers, returnSignature, new String[] {}, new String[] {});
    }

    public Method(Class declaringClass, String name, int modifiers, String returnSignature, String[] parameterSignatures) {
        this(declaringClass, name, modifiers, returnSignature, parameterSignatures, new String[] {});
    }

    public Method(Class declaringClass, String name, int modifiers, String returnSignature,
            String[] parameterSignatures, String[] exceptionSignatures) {
        this.declaringClass = declaringClass;
        this.name = name;
        this.modifiers = modifiers;
        this.returnSignature = returnSignature;
        this.parameterSignatures = parameterSignatures;
        this.exceptionSignatures = exceptionSignatures;
    }

    /**
     * Returns the Class object representing the class or interface that declares the member or constructor represented
     * by this Member.
     * 
     * @gwtr.progress 0%
     * @return an object representing the declaring class of the underlying member
     */
    public Class getDeclaringClass() {
        return declaringClass;
    }

    /**
     * Returns an array of Class objects that represent the types of the exceptions declared to be thrown by the
     * underlying method represented by this Method object.
     * 
     * @gwtr.progress 0%
     * @return the exception types declared as being thrown by the method this object represents
     */
    public Class[] getExceptionTypes() {
        return (null == exceptionTypes) ? exceptionTypes = GWTRUtil.toTypes(exceptionSignatures) : exceptionTypes;
    }

    /**
     * Returns an array of signatures of Class objects that represent the types of the exceptions declared to be thrown
     * by the underlying method represented by this Method object.
     * 
     * @return the exception types declared as being thrown by the method this object represents
     */
    public String[] getExceptionSignatures() {
        return exceptionSignatures;
    }

    /**
     * Returns the Java language modifiers for the method represented by this Method object, as an integer.
     * 
     * @gwtr.progress 0%
     * @return the Java language modifiers for the underlying member
     */
    public int getModifiers() {
        return modifiers;
    }

    /**
     * Returns the name of the method represented by this Method object, as a String.
     * 
     * @gwtr.progress 0%
     * @return the simple name of the underlying member
     */
    public String getName() {
        return name;
    }

    /**
     * Returns an array of Class objects that represent the formal parameter types, in declaration order, of the method
     * represented by this Method object. Returns an array of length 0 if the underlying method takes no parameters.
     * 
     * @gwtr.progress 0%
     * @return
     */
    public Class[] getParameterTypes() {
        return (null == parameterTypes) ? parameterTypes = GWTRUtil.toTypes(parameterSignatures) : parameterTypes;
    }

    /**
     * Returns a Class object that represents the formal return type of the method represented by this Method object.
     * 
     * @gwtr.progress 0%
     * @return the return type for the method this object represents
     */
    public Class getReturnType() {
        return (null == returnType) ? returnType = GWTRUtil.toType(returnSignature) : returnType;
    }

    /**
     * Returns a signature of Class object that represents the formal return type of the method represented by this
     * Method object.
     * 
     * @return the return signature of type for the method this object represents
     */
    public String getReturnSignature() {
        return returnSignature;
    }

    /**
     * Returns the name of the method represented by this Method object witch signatures of underlying method parameters
     * 
     * @return full method name with signatures as example: myMethod(Ljava.lang.Integer;IJ);
     */
    public String getSignaturedName() {
        return GWTRUtil.toMethodSignaturedName(getName(), getParameterSignateresString());
    }

    public String[] getParameterSignatures() {
        return parameterSignatures;
    }

    public String getParameterSignateresString() {
        return GWTRUtil.toSignatureString(parameterSignatures);
    }

    /**
     * Invokes the underlying method represented by this Method object, on the specified object with the specified
     * parameters. Individual parameters are automatically unwrapped to match primitive formal parameters, and both
     * primitive and reference parameters are subject to method invocation conversions as necessary.
     * <p>
     * If the underlying method is static, then the specified obj argument is ignored. It may be null.
     * <p>
     * If the number of formal parameters required by the underlying method is 0, the supplied args array may be of
     * length 0 or null.
     * <p>
     * If the underlying method is an instance method, it is invoked using dynamic method lookup as documented in The
     * Java Language Specification, Second Edition, section 15.12.4.4; in particular, overriding based on the runtime
     * type of the target object will occur.
     * <p>
     * If the underlying method is static, the class that declared the method is initialized if it has not already been
     * initialized.
     * <p>
     * If the method completes normally, the value it returns is returned to the caller of invoke; if the value has a
     * primitive type, it is first appropriately wrapped in an object. If the underlying method return type is void, the
     * invocation returns null.
     * 
     * @param obj
     *            the object the underlying method is invoked from
     * @param args
     *            the arguments used for the method call
     * @return the result of dispatching the method represented by this object on obj with parameters args
     * @throws IllegalArgumentException,InvocationTargetException 
     * @throws IllegalAccessException
     *             if this Method object enforces Java language access control and the underlying method is
     *             inaccessible.
     * @throws IllegalArgumentException
     *             if the method is an instance method and the specified object argument is not an instance of the class
     *             or interface declaring the underlying method (or of a subclass or implementor thereof); if the number
     *             of actual and formal parameters differ; if an unwrapping conversion for primitive arguments fails; or
     *             if, after possible unwrapping, a parameter value cannot be converted to the corresponding formal
     *             parameter type by a method invocation conversion.
     * @throws InvocationTargetException
     *             if the underlying method throws an exception.
     * @throws NullPointerException
     *             if the specified object is null and the method is an instance method.
     * @throws ExceptionInInitializerError
     *             if the initialization provoked by this method fails.
     */
    public Object invoke(Object obj, Object[] args) 
    throws IllegalArgumentException,
    InvocationTargetException,IllegalAccessException,InvocationTargetException,NullPointerException {
        // TODO : make it more smooth...
        // TODO : add throws
        try {
            if (args != null && parameterSignatures != null && args.length == parameterSignatures.length)
                return nativeInvoke(obj, args == null ? null : JavaToJSArrayUtil.arrayConvert(args, parameterSignatures));
            else if ((args == null && parameterSignatures == null)
                    || (args == null && parameterSignatures.length == 0)) {
                return nativeInvoke(obj,  null);
    
            } else
                throw new IllegalArgumentException();
        } catch (Exception ex) {
            throw new InvocationTargetException(ex.getCause());
        } 
    }

}
